#! /usr/bin/env python
'''
Handle Nintendo Wii DVD file system
'''

# NiemaFS imports
from niemafs.common import clean_string, FileSystem, safename
from niemafs.gcm import GcmFS

# imports
from Crypto.Cipher import AES
from datetime import datetime
from multiprocessing import Pool
from pathlib import Path
from struct import unpack
from warnings import warn

# constants
COMMON_KEY = [
    b"\xEB\xE4\x2A\x22\x5E\x85\x93\xE4\x48\xD9\xC5\x45\x73\x81\xAA\xF7", # Common Key: https://github.com/grp/Wii.py/blob/6978355656ac73aa73ef1eea0f52c50644010b56/title.py#L76
    b"\x63\xB8\x2B\xB4\xF4\x61\x4E\x2E\x13\xF2\xFE\xFB\xBA\x4C\x9B\x7E", # Korean Key: https://github.com/grp/Wii.py/blob/6978355656ac73aa73ef1eea0f52c50644010b56/title.py#L77
]
PARTITION_TYPE = [
    'data',              # type 0
    'update',            # type 1
    'channel_installer', # type 2
]
REGION = {
    0: 'Japan/Taiwan',
    1: 'USA',
    2: 'PAL',
    4: 'Korea',
}

# helper function for multiprocessing of AES decryption
def aes_helper(aes_tup):
    decrypt_key, cluster_data_encrypted = aes_tup
    aes = AES.new(decrypt_key, AES.MODE_CBC, cluster_data_encrypted[0x3D0 : 0x3E0])
    return aes.decrypt(cluster_data_encrypted[0x0400:])

class WiiFS(FileSystem):
    '''Class to represent a `Nintendo Wii DVD <https://wiibrew.org/wiki/Wii_disc#%22System_Area%22>`_.'''
    def __init__(self, file_obj, path=None):
        # set things up
        if file_obj is None:
            raise ValueError("file_obj must be a file-like")
        super().__init__(path=path, file_obj=file_obj)
        self.header = None           # Header
        self.volume_info = None      # Volume (Partitions) Information
        self.partition_tables = None # Partition Tables
        self.region_info = None      # Region Information

    def get_header(self):
        '''Return the `Header <https://wiibrew.org/wiki/Wii_disc#Header>`_ of the Wii disc.

        Returns:
            `bytes`: The Header of the Wii disc.
        '''
        if self.header is None:
            self.header = self.read_file(0x0000, 0x0400)
        return self.header

    def get_volume_info(self):
        '''Return the `Volume (Partitions) Information <https://wiibrew.org/wiki/Wii_disc#Partitions_information>`_ of the Wii disc.

        Returns:
            `bytes`: The Volume (Partitions) Information of the Wii disc.
        '''
        if self.volume_info is None:
            self.volume_info = self.read_file(0x40000, 32)
        return self.volume_info

    def get_partition_tables(self):
        '''Return the `Partition Tables <https://wiibrew.org/wiki/Wii_disc#Partition_table_entry>`_ of the Wii disc.

        Returns:
            `list` of `bytes`: The Partition Tables of the Wii disc.
        '''
        if self.partition_tables is None:
            self.partition_tables = [self.read_file(parts_info['table_offset'], 8 * parts_info['num_partitions']) for parts_info in self.parse_volume_info()]
        return self.partition_tables

    def get_region_info(self):
        '''Return the `Region Information <https://wiibrew.org/wiki/Wii_disc#Region_setting>`_ of the Wii disc.

        Returns:
            `bytes`: The Region Information of the Wii disc.
        '''
        if self.region_info is None:
            self.region_info = self.read_file(0x4E000, 32)
        return self.region_info

    def parse_header(data):
        '''Return a parsed version of the `Header <https://wiibrew.org/wiki/Wii_disc#Header>`_ of the Wii disc or partition.

        Returns:
            `dict`: A parsed version of the Header of the Wii disc or partition.
        '''
        # set things up
        out = dict()

        # parse raw Header data
        out['game_code'] =                 data[0x0000 : 0x0004] # Game Code "XYYZ": X = Console ID, YY = Game Code, Z = Country Code
        out['maker_code'] =                data[0x0004 : 0x0006] # Maker Code
        out['disk_id'] =                   data[0x0006]          # Disk ID
        out['version'] =                   data[0x0007]          # Version
        out['audio_streaming'] =           data[0x0008]          # Audio Streaming
        out['stream_buffer_size'] =        data[0x0009]          # Stream Buffer Size
        out['offsets_0x000A_0x0017'] =     data[0x000A : 0x0018] # Unused (should be 0s)
        out['wii_magic_word'] =            data[0x0018 : 0x001C] # Wii Magic Word (should be 0x5d1c9ea3)
        out['gc_magic_word'] =             data[0x001C : 0x0020] # GameCube Magic Word (should be 0s on Wii discs)
        out['game_name'] =                 data[0x0020 : 0x0060] # Game Name
        out['disable_hash_verification'] = data[0x0060]          # Disable Hash Verification
        out['disable_disc_encryption'] =   data[0x0061]          # Disable Disc Encryption and H3 Hash Table Load/Verification
        out['offsets_0x0062_0x007F'] =     data[0x0062 : 0x0080] # Unknown
        out['offsets_0x0080_0x043F'] =     data[0x0080 : 0x0440] # Unushed (should be 0s)

        # clean strings
        for k in ['game_code', 'maker_code', 'game_name']:
            try:
                out[k] = clean_string(out[k])
            except:
                warn("Unable to parse Header '%s' as string: %s" % (k, out[k]))

        # return final parsed data
        return out

    def parse_volume_info(self):
        '''Return a parsed version of the `Volume (Partitions) Information <https://wiibrew.org/wiki/Wii_disc#Partitions_information>`_ of the Wii disc.

        Returns:
            `list` of `dict`: A parsed version of the Volume (Partitions) Information of the Wii disc.
        '''
        # set things up
        data = self.get_volume_info()
        out = [dict(), dict(), dict(), dict()]

        # parse raw Volume (Partitions) Information data
        out[0]['num_partitions'] = unpack('>I', data[0 : 4])[0]        # Number of Volume 0 Partitions
        out[0]['table_offset'] =   unpack('>I', data[4 : 8])[0]   << 2 # Volume 0 Info Table Offset
        out[1]['num_partitions'] = unpack('>I', data[8 : 12])[0]       # Number of Volume 1 Partitions
        out[1]['table_offset'] =   unpack('>I', data[12 : 16])[0] << 2 # Volum2 1 Info Table Offset
        out[2]['num_partitions'] = unpack('>I', data[16 : 20])[0]      # Number of Volume 2 Partitions
        out[2]['table_offset'] =   unpack('>I', data[20 : 24])[0] << 2 # Volume 2 Info Table Offset
        out[3]['num_partitions'] = unpack('>I', data[24 : 28])[0]      # Number of Volume 3 Partitions
        out[3]['table_offset'] =   unpack('>I', data[28 : 32])[0] << 2 # Volume 3 Info Table Offset

        # return final parsed data
        return out

    def parse_partition_tables(self):
        '''Return a parsed version of the `Partition Tables <https://wiibrew.org/wiki/Wii_disc#Partition_table_entry>`_ of the Wii disc.

        Returns:
            `list` of `dict`: A parsed version of the Partition Tables
        '''
        # set things up
        raw_tables = self.get_partition_tables()
        out = list()

        # parse raw Partition Table data
        for table_num, data in enumerate(raw_tables):
            partitions = list()
            for i in range(0, len(data), 8):
                partitions.append({
                    'offset': unpack('>I', data[i : i + 4])[0] << 2, # Partition Offset
                    'type':   unpack('>I', data[i + 4 : i + 8])[0],  # Partition Type (0 = Data Partition, 1 = Update Partition, 2 = Channel Installer)
                })
            out.append(partitions)
        
        # return final parsed data
        return out

    def parse_region_info(self):
        '''Return a parsed version of the `Region Information <https://wiibrew.org/wiki/Wii_disc#Region_setting>`_ of the Wii disc.

        Returns:
            `dict`: A parsed version of the Region Information of the Wii disc.
        '''
        # set things up
        data = self.get_region_info()
        out = dict()

        # parse raw Region Information data
        out['region'] =                  unpack('>I', data[0x00 : 0x04])[0] # Region (0 = Japan/Taiwan, 1 = USA, 2 = PAL, 4 = Korea)
        out['offsets_0x04_0x0F'] =       data[0x04 : 0x10]                  # Unused(?)
        out['age_rating_japan_taiwan'] = data[0x10]                         # Age Rating: Japan/Taiwan
        out['age_rating_usa'] =          data[0x11]                         # Age Rating: USA
        out['offset_0x12'] =             data[0x12]                         # Unused(?)
        out['age_rating_germany'] =      data[0x13]                         # Age Rating: Germany
        out['age_rating_pegi'] =         data[0x14]                         # Age Rating: PEGI
        out['age_rating_finland'] =      data[0x15]                         # Age Rating: Finland
        out['age_rating_portugal'] =     data[0x16]                         # Age Rating: Portugal
        out['age_rating_britain'] =      data[0x17]                         # Age Rating: Britain
        out['age_rating_australia'] =    data[0x18]                         # Age Rating: Australia
        out['age_rating_korea'] =        data[0x19]                         # Age Rating: Korea
        out['offsets_0x1A_0x1F'] =       data[0x1A : 0x20]                  # Unused(?)

        # parse region byte
        try:
            out['region'] = REGION[out['region']]
        except:
            warn("Unable to parse region byte: %s" % out['region'])

        # return final parsed data
        return out

    def parse_ticket(data):
        '''Return a parsed version of a `Ticket <https://wiibrew.org/wiki/Ticket>`_ of the Wii disc.

        Args:
            `data` (`bytes`): The raw Ticket data.

        Returns:
            `dict`: A parsed version of a Ticket of the Wii disc.
        '''
        # set things up
        out = dict()

        # parse Signed Blob Header: https://wiibrew.org/wiki/Ticket#Signed_blob_header
        out['signature_type'] =        data[0x0000 : 0x0004] # Signature Type (0x10001 for RSA-2048)
        out['signature_by_cert_key'] = data[0x0004 : 0x0104] # Signature by a Certificate's Key
        out['padding_module_64'] =     data[0x0104 : 0x0140] # Padding Module 64

        # parse v0 Ticket: https://wiibrew.org/wiki/Ticket#Signed_blob_header
        out['signature_issuer'] =           data[0x0140 : 0x0180]                  # Signature Issuer
        out['ecdh'] =                       data[0x0180 : 0x01BC]                  # ECDH Data (used to generate one-time key during install of console specific titles)
        out['ticket_format_version'] =      data[0x01BC]                           # Ticket Format Version
        out['offsets_0x01BD_0x01BE'] =      data[0x01BD : 0x01BF]                  # Reserved
        out['title_key'] =                  data[0x01BF : 0x01CF]                  # Title Key (encrypted by Common Key)
        out['offset_0x01CF'] =              data[0x01CF]                           # Unknown
        out['ticket_id'] =                  data[0x01D0 : 0x01D8]                  # Ticket ID (used as Initialization Vector (IV) for title key decryption of console specific titles)
        out['console_id'] =                 data[0x01D8 : 0x01DC]                  # Console ID (NG ID in console specific titles: https://wiibrew.org/wiki/Hardware/OTP)
        out['title_id'] =                   data[0x01DC : 0x01E4]                  # Title ID (used as Initialization Vector (IV) for AES-CBC encryption)
        out['offsets_0x01E4_0x01E5'] =      data[0x01E4 : 0x01E6]                  # Unknown (usually 0xFFFF)
        out['ticket_title_version'] =       unpack('>H', data[0x01E6 : 0x01E8])[0] # Ticket Title Version
        out['permitted_titles_mask'] =      data[0x01E8 : 0x01EC]                  # Permitted Titles Mask
        out['permit_mask'] =                data[0x01EC : 0x01F0]                  # Permit Mask (disk title is ANDed with inverse of this mask to see if result matches Permitted Titles Mask)
        out['title_export_allowed'] =       data[0x01F0]                           # Title Export allowed using PRNG key (1 = allowed, 0 = not allowed)
        out['common_key_index'] =           data[0x01F1]                           # Common Key Index (0 = Common Key, 1 = Korean Key, 2 = Wii U Wii Mode)
        out['offsets_0x01F2_0x01F4'] =      data[0x01F2 : 0x01F5]                  # Unknown
        out['offsets_0x01F5_0x0221'] =      data[0x01F5 : 0x0222]                  # Unknown
        out['content_access_permissions'] = data[0x0222 : 0x0262]                  # Content Access Permissions (1 bit for each content)
        out['offsets_0x0262_0x0263'] =      data[0x0262 : 0x0264]                  # Padding (always 0s)
        out['limit_type_1'] =               unpack('>I', data[0x0264 : 0x0268])[0] # Limit Type 1 (0 = disable, 1 = time limit (minutes), 3 = disable, 4 = launch count limit)
        out['max_usage_1'] =                unpack('>I', data[0x0268 : 0x026C])[0] # Maximum Usage 1 (depending on limit type)
        out['limit_type_2'] =               unpack('>I', data[0x026C : 0x0270])[0] # Limit Type 2
        out['max_usage_2'] =                unpack('>I', data[0x0270 : 0x0274])[0] # Maximum Usage 2
        out['limit_type_3'] =               unpack('>I', data[0x0274 : 0x0278])[0] # Limit Type 3
        out['max_usage_3'] =                unpack('>I', data[0x0278 : 0x027C])[0] # Maximum Usage 3
        out['limit_type_4'] =               unpack('>I', data[0x027C : 0x0280])[0] # Limit Type 4
        out['max_usage_4'] =                unpack('>I', data[0x0280 : 0x0284])[0] # Maximum Usage 4
        out['limit_type_5'] =               unpack('>I', data[0x0284 : 0x0288])[0] # Limit Type 5
        out['max_usage_5'] =                unpack('>I', data[0x0288 : 0x028C])[0] # Maximum Usage 5
        out['limit_type_6'] =               unpack('>I', data[0x028C : 0x0290])[0] # Limit Type 6
        out['max_usage_6'] =                unpack('>I', data[0x0290 : 0x0294])[0] # Maximum Usage 6
        out['limit_type_7'] =               unpack('>I', data[0x0294 : 0x0298])[0] # Limit Type 7
        out['max_usage_7'] =                unpack('>I', data[0x0298 : 0x029C])[0] # Maximum Usage 7
        out['limit_type_8'] =               unpack('>I', data[0x029C : 0x02A0])[0] # Limit Type 8
        out['max_usage_8'] =                unpack('>I', data[0x02A0 : 0x02A4])[0] # Maximum Usage 8

        # parse v1 Ticket: https://wiibrew.org/wiki/Ticket#v1_ticket
        if len(data) > 0x02A4:
            out['v1_header_version'] =         unpack('>H', data[0x02A4 : 0x02A6])[0] # v1 Ticket Header Version
            out['v1_header_size'] =            unpack('>H', data[0x02A6 : 0x02A8])[0] # v1 Ticket Header Size
            out['v1_ticket_size'] =            unpack('>I', data[0x02A8 : 0x02AC])[0] # v1 Ticket Size (0x14?)
            out['v1_section_headers_offset'] = unpack('>I', data[0x02AC : 0x02B0])[0] # v1 Offset of Section Headers
            out['v1_section_headers_count'] =  unpack('>H', data[0x02B0 : 0x02B2])[0] # v1 Number of Section Headers
            out['v1_section_headers_size'] =   unpack('>H', data[0x02B2 : 0x02B4])[0] # v1 Size of Each Section Header (0x14?)
            out['v1_misc_flags'] =             data[0x02B4 : 0x02B8]                  # v1 Miscellaneous Flags

        # clean strings
        for k in ['signature_issuer']:
            try:
                out[k] = clean_string(out[k])
            except:
                warn("Unable to parse Ticket '%s' as string: %s" % (k, out[k]))

        # return final parsed data
        return out

    def __iter__(self):
        # parse each partition table
        for volume_num, parsed_partition_table in enumerate(self.parse_partition_tables()):
            volume_path = Path('Volume %d' % volume_num)
            yield (volume_path, None, None)

            # parse each partition in the partition table
            for partition_num, partition in enumerate(parsed_partition_table):
                # load partition header
                try:
                    partition_type = PARTITION_TYPE[partition['type']]
                except:
                    partition_type = 'type%d' % partition['type']
                partition_header = self.read_file(partition['offset'], 0x02C0)

                # parse partition header: ticket
                ticket = partition_header[0x0000 : 0x02A4]
                parsed_ticket = WiiFS.parse_ticket(ticket)
                aes = AES.new(COMMON_KEY[parsed_ticket['common_key_index']], AES.MODE_CBC, parsed_ticket['title_id'] + (b'\x00'*8))
                decrypt_key = aes.decrypt(parsed_ticket['title_key'])

                # parse partition header: TMD
                tmd_size =   unpack('>I', partition_header[0x02A4 : 0x02A8])[0]
                tmd_offset = unpack('>I', partition_header[0x02A8 : 0x02AC])[0] << 2
                tmd = self.read_file(tmd_offset, tmd_size)

                # parse partition header: cert chain
                cert_chain_size =   unpack('>I', partition_header[0x02AC : 0x02B0])[0]
                cert_chain_offset = unpack('>I', partition_header[0x02B0 : 0x02B4])[0] << 2
                cert_chain = self.read_file(cert_chain_offset, cert_chain_size)

                # parse partition header: H3 table
                h3_table_offset = unpack('>I', partition_header[0x02B4 : 0x02B8])[0] << 2 # size is always 0x18000
                h3_table = self.read_file(h3_table_offset, 0x18000)

                # parse partition header: partition data location
                partition_data_offset = unpack('>I', partition_header[0x02B8 : 0x02BC])[0] << 2
                partition_data_size =   unpack('>I', partition_header[0x02BC : 0x02C0])[0] << 2

                # decrypt partition data: https://wiibrew.org/wiki/Wii_disc#Partition_Data
                data_start_offset = partition['offset'] + partition_data_offset
                data_end_offset = data_start_offset + partition_data_size
                if WiiFS.parse_header(self.get_header())['disable_disc_encryption'] == 0:
                    with Pool(processes=None) as pool:
                        data_decrypted = b''.join(pool.map(aes_helper, [(decrypt_key, self.read_file(cluster_offset, 0x8000)) for cluster_offset in range(data_start_offset, data_end_offset, 0x8000)]))
                else:
                    data_decrypted = self.read_file(data_start_offset, partition_data_size)

                # yield partition header data
                partition_header = WiiFS.parse_header(data_decrypted[0x0000 : 0x0420])
                partition_path = volume_path / ('Partition %d (%s) (%s%s) (%s)' % (partition_num, partition_type.capitalize(), partition_header['game_code'], partition_header['maker_code'], safename(partition_header['game_name'])))
                yield (partition_path, None, None)
                partition_header_path = partition_path / '___PARTITION_HEADER_NIEMAFS'
                yield (partition_header_path, None, None)
                yield (partition_header_path / 'ticket.bin', None, ticket)
                yield (partition_header_path / 'tmd.bin', None, tmd)
                yield (partition_header_path / 'cert.bin', None, cert_chain)
                yield (partition_header_path / 'h3.bin', None, h3_table)

                # parse decrypted data, similar to GameCube: https://github.com/niemasd/NiemaFS/blob/main/niemafs/gcm.py
                fst_offset = unpack('>I', data_decrypted[0x0424 : 0x0428])[0] << 2
                num_entries = unpack('>I', data_decrypted[fst_offset + 8 : fst_offset + 12])[0]
                string_table_start = fst_offset + (12 * num_entries)
                string_table_end = string_table_start
                for _ in range(num_entries - 1):
                    string_table_end = data_decrypted.index(b'\00', string_table_end + 1)
                fst = data_decrypted[fst_offset : string_table_end + 1]
                to_visit = [GcmFS.parse_fst(0, None, fst, string_table_start - fst_offset)] # start at root directory
                while len(to_visit) != 0:
                    file_entry = to_visit.pop()
                    if len(file_entry['children']) == 0:
                        file_entry['offset'] <<= 2 # GameCube offsets are exact, Wii offsets are >> 2
                        file_data = data_decrypted[file_entry['offset'] : file_entry['offset'] + file_entry['length']]
                    else:
                        file_data = None
                        to_visit += file_entry['children']
                    if not file_entry['is_root']:
                        yield (partition_path / file_entry['path'], None, file_data)



#! /usr/bin/env python
'''
Handle ISO 9660 file systems
'''

# NiemaFS imports
from niemafs.common import clean_string, FileSystem

# imports
from datetime import datetime
from math import ceil
from pathlib import Path
from struct import unpack
from warnings import warn

# constants
ISO9660_PVD_MAGIC_WORD = bytes(ord(c) for c in 'CD001')  # https://wiki.osdev.org/ISO_9660#Volume_Descriptors
MAGIC_WORD_SEARCH_SIZE = 50000  # fallback search window if auto-detect fails

# Common optical disc sector layouts (physical sector bytes, user data offset, user data size)
COMMON_LAYOUT_CANDIDATES = [
    (2048,   0, 2048), # Standard ISO/UDF logical blocks
    (2352,  16, 2048), # CD-ROM Mode 1 raw: sync(12)+hdr(4)+data(2048)+...
    (2352,  16, 2336), # CD-ROM Mode 2 raw "formless" data payload after hdr (rare for ISO9660)
    (2352,  24, 2048), # CD-ROM XA Mode 2 Form 1 raw: +subhdr(8) then 2048
    (2352,  24, 2324), # CD-ROM XA Mode 2 Form 2 raw: +subhdr(8) then 2324
    (2352,   0, 2352), # CD-DA audio frame (Red Book)
    (2448,  16, 2048), # Mode 1 raw + 96B subchannel
    (2448,  16, 2336), # Mode 2 raw + 96B subchannel (formless)
    (2448,  24, 2048), # XA Form 1 raw + 96B subchannel
    (2448,  24, 2324), # XA Form 2 raw + 96B subchannel
    (2448,   0, 2352), # Audio + 96B subchannel (CDG/Karaoke/CD-DA+sub)
    (2340,   4, 2048), # Mode 1 without sync (hdr at 0..3, user at 4)
    (2340,   4, 2336), # Mode 2 without sync (hdr at 0..3, "data" after)
    (2336,   0, 2336), # Mode 2 payload-only (no sync/hdr)
    (2336,   0, 2048), # Mode 1 sans sync+hdr: first 2048 are user, then EDC/zero/ECC
    (2324,   0, 2324), # XA Form 2 user data only
    (2328,   0, 2324), # XA Form 2 "data + 4 spare" (libcdio notes 2328 = 2324+4 spare)
    (2332,   8, 2324), # XA "subheader + 2324" (8 + 2324) a.k.a. M2SUB_SECTOR_SIZE
    (2056,   8, 2048), # XA Form 1 cooked with subheader preserved (8 + 2048)
    (2052,   0, 2048), # Mode 1 cooked-ish with EDC preserved (2048 + 4)
    (2064,  16, 2048), # Mode 1 with only sync+hdr kept (12+4+2048)
    (2076,  16, 2048), # Mode 1 with sync+hdr+data+EDC+zero (12+4+2048+4+8)
]

class IsoFS(FileSystem):
    '''Class to represent an `ISO 9660 <https://wiki.osdev.org/ISO_9660>`_ optical disc'''
    def __init__(self, file_obj, path=None):
        # set things up
        if file_obj is None:
            raise ValueError("file_obj must be a file-like")
        super().__init__(path=path, file_obj=file_obj)
        self.logical_block_size = None           # ISO logical block size (usually 2048)
        self.physical_logical_block_size = None  # bytes per physical sector in the image (e.g., 2048, 2352, 2448)
        self.user_data_offset = None             # byte offset of ISO user data within a physical sector
        self.user_data_size = None               # bytes of ISO user data per physical sector (usually 2048)
        self.system_area = None                  # system area data
        self.volume_descriptors = dict()         # keys = Volume Descriptor Type codes, values = bytes

        # detect sector layout and load header to ensure file validity up-front
        self.detect_layout()
        self.get_system_area()
        self.get_volume_descriptors()

        # if the PVD states a logical block size, honor it (it should be 2048 for ISO 9660)
        try:
            pvd = self.parse_primary_volume_descriptor()
            if pvd is not None:
                lbs = pvd.get('logical_block_size_LE')
                if isinstance(lbs, int) and lbs > 0:
                    self.logical_block_size = lbs
        except:
            pass # if PVD parsing fails here, the image might still be readable via other volume descriptors.

    def tz_offset_to_datetime_str(x):
        '''Convert an ISO 9660 timezone offset to a `datetime` format string

        Args:
            `x` (`int`): The ISO 9660 timezone offset.

        Returns:
            `str`: The `datetime` format string.
        '''
        tz_offset_hours = (x / 4) - 12
        tz_offset_sign = '-' if tz_offset_hours < 0 else '+'
        tz_offset_hh = str(abs(int(tz_offset_hours))).zfill(2)
        tz_offset_mm = str(int((abs(tz_offset_hours) % 1) * 60)).zfill(2)
        return '%s%s%s' % (tz_offset_sign, tz_offset_hh, tz_offset_mm)

    def parse_pvd_datetime(data):
        '''Parse a date/time in the `ISO 9660 Primary Volume Descriptor (PVD) date/time format <https://wiki.osdev.org/ISO_9660#Date/time_format>`_.

        Args:
            `data` (`bytes`): A date/time (exactly 17 bytes) in the ISO 9660 PVD date/time format.

        Returns:
            `datetime`: A Python `datetime` object.
        '''
        if len(data) != 17:
            raise ValueError("ISO 9660 PVD date/time must be exactly 17 bytes: %s" % data)
        dt_str = ''.join(str(v) if v < 48 else chr(v) for v in data[0:16]) + '0000' # chr(48) == '0'
        tz_offset_str = IsoFS.tz_offset_to_datetime_str(data[16])
        try:
            return datetime.strptime(dt_str + tz_offset_str, '%Y%m%d%H%M%S%f%z')
        except ValueError:
            return datetime.strptime(dt_str, '%Y%m%d%H%M%S%f') # timezone is sometimes messed up

    def parse_directory_datetime(data):
        '''Parse a date/time in the `ISO 9660 directory record date/time format <https://wiki.osdev.org/ISO_9660#Directories>.`_

        Args:
            `data` (`bytes`): A date/time (exactly 7 bytes) in the ISO 9660 directory record date/time format.

        Returns:
            `datetime`: A Python `datetime` object.
        '''
        dt_str = str(data[0] + 1900) + ''.join(str(x).zfill(2) for x in data[1:6])
        tz_offset_str = IsoFS.tz_offset_to_datetime_str(data[6])
        try:
            return datetime.strptime(dt_str + tz_offset_str, '%Y%m%d%H%M%S%f%z')
        except ValueError:
            return datetime.strptime(dt_str, '%Y%m%d%H%M%S%f') # timezone is sometimes messed up

    def parse_directory_record(data):
        '''Parse an `ISO 9660 directory record <https://wiki.osdev.org/ISO_9660#Directories>`_.

        Args:
            `data` (`bytes`): The raw bytes of the directory record.

        Returns:
            `dict`: The parsed directory record.
        '''
        # parse directory record data
        out = dict()
        out['directory_record_length'] =          data[0]                              # should be equal to len(data)
        out['extended_attribute_record_length'] = data[1]                              # extended attribute record length
        out['data_location_LE'] =                 unpack('<I', data[2:6])[0]           # location (LBA) of data (little-endian)
        out['data_location_BE'] =                 unpack('>I', data[6:10])[0]          # location (LBA) of data (big-endian) (should be equal to previous)
        out['data_length_LE'] =                   unpack('<I', data[10:14])[0]         # length of data (little-endian)
        out['data_length_BE'] =                   unpack('>I', data[14:18])[0]         # length of data (big-endian) (should be equal to previous)
        out['datetime'] =                         data[18:25]                          # recording date and time
        out['file_flags'] =                       data[25]                             # file flags
        out['interleave_file_unit_size'] =        data[26]                             # file unit size for files recorded in interleaved mode (otherwise 0)
        out['interleave_gap_size'] =              data[27]                             # gap size for files recorded in interleaved mode (otherwise 0)
        out['volume_sequence_number_LE'] =        unpack('<H', data[28:30])[0]         # volume this extent is recorded on (little-endian)
        out['volume_sequence_number_BE'] =        unpack('>H', data[30:32])[0]         # volume this extent is recorded on (big-endian) (should be equal to previous)
        out['filename_length'] =                  data[32]                             # length of filename (terminated with ';1' where 1 is file version number)
        out['filename'] =                         data[33 : 33+out['filename_length']] # filename (terminated with ';1' where 1 is file version number)
        # I'm skipping "padding field" and "system use" since they're not useful to me and non-trivial to code

        # parse file flags
        out['file_flags'] = {
            'is_hidden':                         bool(out['file_flags'] & 0b00000001),
            'is_directory':                      bool(out['file_flags'] & 0b00000010),
            'is_associated_file':                bool(out['file_flags'] & 0b00000100),
            'format_in_extended_attribute':      bool(out['file_flags'] & 0b00001000),
            'permissions_in_extended_attribute': bool(out['file_flags'] & 0b00010000),
            'reserved_5':                        bool(out['file_flags'] & 0b00100000),
            'reserved_6':                        bool(out['file_flags'] & 0b01000000),
            'not_final_directory':               bool(out['file_flags'] & 0b10000000),
        }

        # clean strings
        for k in ['filename']:
            try:
                out[k] = clean_string(out[k])
            except:
                warn("Unable to parse Directory Record '%s' as string: %s" % (k, out[k]))

        # parse date-times
        for k in ['datetime']:
            try:
                out[k] = IsoFS.parse_directory_datetime(out[k])
            except:
                warn("Unable to parse Directory Record '%s' as date-time: %s" % (k, out[k]))

        # return final parsed data
        return out

    def read_user_blocks(self, lba, count=1):
        '''Read ISO logical blocks (user data blocks) starting at a specific LBA, returning concatenated user data.

        Args:
            `lba` (`int`): The first LBA of the read.

            `count`: The number of ISO logical blocks to read.

        Returns:
            `bytes`: The read data.
        '''
        if count <= 0:
            return b''
        out = bytearray()
        for i in range(count):
            phys_off = (lba + i) * self.physical_logical_block_size + self.user_data_offset
            out.extend(self.read_file(phys_off, self.user_data_size))
        return bytes(out)

    def read_extent(self, lba, length):
        '''Read bytes from the ISO extent starting at a specific LBA (in user-data LBAs).

        Args:
            `lba`: The first LBA of the read.

            `length` (`int`): The number of bytes to read.

        Returns:
            `bytes`: The read data.
        '''
        if length <= 0:
            return b''
        blocks = ceil(length / self.user_data_size)
        data = self.read_user_blocks(lba, blocks)
        return data[:length]

    def looks_like_pvd(self, block: bytes) -> bool:
        '''Validate the start of an ISO 9660 Volume Descriptor block.

        Args:
            `block` (`bytes`): The block to validate.

        Returns:
            `bool`: `True` if the block looks valid, otherwise `False`.
        '''
        return (block is not None) and (len(block) > 6) and (block[0] == 1) and (block[1:6] == ISO9660_PVD_MAGIC_WORD) and (block[6] == 1)

    def detect_layout(self):
        '''Detect physical sector size, user data offset, and user data size by validating the PVD at LBA 16.'''
        if self.physical_logical_block_size is not None:
            return

        # try common known layouts by reading LBA 16 (PVD location)
        for (phys, off, udsz) in COMMON_LAYOUT_CANDIDATES:
            try:
                self.physical_logical_block_size = phys
                self.user_data_offset = off
                self.user_data_size = udsz
                self.logical_block_size = udsz # logical_block_size == logical block size for ISO-level parsing
                pvd = self.read_user_blocks(16, 1)
                if self.looks_like_pvd(pvd):
                    return
            except:
                pass

        # if we reach here, we failed (unknown layout)
        raise ValueError("ISO layout does not match known existing layouts")

    def get_physical_logical_block_size(self):
        '''Return the ISO physical logical block size.

        Returns:
            `int`: The ISO physical logical block size.
        '''
        if self.physical_logical_block_size is None:
            self.detect_layout()
        return self.physical_logical_block_size

    def get_user_data_offset(self):
        '''Return the ISO user data offset.

        Returns:
            `int`: The ISO user data offset.
        '''
        if self.user_data_offset is None:
            self.detect_layout()
        return self.user_data_offset

    def get_user_data_size(self):
        '''Return the ISO user data size.

        Returns:
            `int`: The ISO user data size.
        '''
        if self.user_data_size is None:
            self.detect_layout()
        return self.user_data_size

    def get_logical_block_size(self):
        '''Return the ISO logical block size.

        Returns:
            `int`: The ISO logical block size in bytes.
        '''
        if self.logical_block_size is None:
            self.detect_layout()
        return self.logical_block_size

    def get_system_area(self):
        '''Return the System Area (logical sectors 0x00-0x0F = first 16 sectors) of the ISO.

        Returns:
            `bytes`: The System Area (first 16 ISO logical blocks).
        '''
        if self.system_area is None:
            self.system_area = self.read_user_blocks(0, 16)
        return self.system_area

    def get_volume_descriptors(self):
        '''Return the Volume Descriptors of the ISO.

        Returns:
            `dict`: Keys are `Volume Descriptor Type codes <https://wiki.osdev.org/ISO_9660#Volume_Descriptor_Type_Codes>`_, and values are `bytes` of the corresponding volume descriptor.
        '''
        if len(self.volume_descriptors) == 0:
            lba = 16 # Volume Descriptors begin at LBA 16 and continue until type code 255
            while True:
                next_volume_descriptor = self.read_user_blocks(lba, 1)
                if len(next_volume_descriptor) < 7 or next_volume_descriptor[1:6] != ISO9660_PVD_MAGIC_WORD:
                    warn("Volume Descriptor at LBA %d does not look like an ISO 9660 descriptor" % lba)
                self.volume_descriptors[next_volume_descriptor[0]] = next_volume_descriptor
                if next_volume_descriptor[0] == 255:  # Volume Descriptor Set Terminator
                    break
                lba += 1
        return self.volume_descriptors

    def get_boot_record(self):
        '''Return the Boot Record (Volume Descriptor code 0) of the ISO.

        Returns:
            `bytes`: The Boot Record (Volume Descriptor code 0) of the ISO, or `None` if the ISO does not have one.
        '''
        try:
            return self.get_volume_descriptors()[0]
        except KeyError:
            return None

    def get_primary_volume_descriptor(self):
        '''Return the Primary Volume Descriptor (PVD; Volume Descriptor code 1) of the ISO.

        Returns:
            `bytes`: The Primary Volume Descriptor (PVD; Volume Descriptor code 1) of the ISO, or `None` if the ISO does not have one.
        '''
        try:
            return self.get_volume_descriptors()[1]
        except KeyError:
            return None

    def get_supplementary_volume_descriptor(self):
        '''Return the Supplementary Volume Descriptor (Volume Descriptor code 2) of the ISO.

        Returns:
            `bytes`: The Supplementary Volume Descriptor (Volume Descriptor code 2) of the ISO, or `None` if the ISO does not have one.
        '''
        try:
            return self.get_volume_descriptors()[2]
        except KeyError:
            return None

    def get_volume_partition_descriptor(self):
        '''Return the Volume Partition Descriptor (Volume Descriptor code 3) of the ISO.

        Returns:
            `bytes`: The Volume Partition Descriptor (Volume Descriptor code 3) of the ISO, or `None` if the ISO does not have one.
        '''
        try:
            return self.get_volume_descriptors()[3]
        except KeyError:
            return None

    def get_volume_descriptor_set_terminator(self):
        '''Return the Volume Descriptor Set Terminator (Volume Descriptor code 0xFF = 255) of the ISO.

        Returns:
            `bytes`: The Volume Descriptor Set Terminator (Volume Descriptor code 0xFF = 255) of the ISO, or `None` if the ISO does not have one.
        '''
        try:
            return self.get_volume_descriptors()[255]
        except KeyError:
            raise ValueError("ISO does not have a Volume Descriptor Set Terminator")

    def parse_boot_record(self):
        '''Return a parsed version of the `Boot Record <https://wiki.osdev.org/ISO_9660#The_Boot_Record>`_ of the ISO.

        Returns:
            `dict`: A parsed version of the Boot Record of the ISO, or `None` if the ISO does not have one.
        '''
        # set things up
        br = self.get_boot_record()
        if br is None:
            return None
        out = dict()

        # parse raw Boot Record data
        out['type_code'] =              br[0]     # should always be 0
        out['identifier'] =             br[1:6]   # should always be "CD001"
        out['version'] =                br[6]     # should always be 1?
        out['boot_system_identifier'] = br[7:39]  # ID of the system which can act on and boot the system from the boot record
        out['boot_identifier'] =        br[39:71] # ID of the boot system defined in the rest of this descriptor
        out['boot_system_use'] =        br[71:]   # Custom - used by the boot system

        # clean strings
        for k in ['identifier', 'boot_system_identifier', 'boot_identifier']:
            try:
                out[k] = clean_string(out[k])
            except:
                warn("Unable to parse Boot Record '%s' as string: %s" % (k, out[k]))

        # return final parsed data
        return out

    def parse_primary_volume_descriptor(self):
        '''Return a parsed version of the `Primary Volume Descriptor (PVD) <https://wiki.osdev.org/ISO_9660#The_Primary_Volume_Descriptor>`_ of the ISO.

        Returns:
            `dict`: A parsed version of the Primary Volume Descriptor (PVD) of the ISO, or `None` if the ISO does not have one.
        '''
        # set things up
        pvd = self.get_primary_volume_descriptor()
        if pvd is None:
            return None
        out = dict()

        # parse raw PVD data
        out['type_code'] =                      pvd[0]                        # should always be 1
        out['identifier'] =                     pvd[1:6]                      # should always be "CD001"
        out['version'] =                        pvd[6]                        # should always be 1?
        out['offset_7'] =                       pvd[7]                        # should always be 0
        out['system_identifier'] =              pvd[8:40]                     # Name of the system that can act upon sectors 0x00-0x0F for the volume
        out['volume_identifier'] =              pvd[40:72]                    # Identification (label) of this volume
        out['offsets_72_79'] =                  pvd[72:80]                    # should always be all 0s
        out['volume_space_size_LE'] =           unpack('<I', pvd[80:84])[0]   # Volume Space Size (little-endian)
        out['volume_space_size_BE'] =           unpack('>I', pvd[84:88])[0]   # Volume Space Size (big-endian) (should be equal to previous)
        out['offsets_88_119'] =                 pvd[88:120]                   # should always be all 0s
        out['volume_set_size_LE'] =             unpack('<H', pvd[120:122])[0] # Volume Set Size (little-endian)
        out['volume_set_size_BE'] =             unpack('>H', pvd[122:124])[0] # Volume Set Size (big-endian) (should be equal to previous)
        out['volume_sequence_number_LE'] =      unpack('<H', pvd[124:126])[0] # Volume Sequence Number (little-endian)
        out['volume_sequence_number_BE'] =      unpack('>H', pvd[126:128])[0] # Volume Sequence Number (big-endian) (should be equal to previous)
        out['logical_block_size_LE'] =          unpack('<H', pvd[128:130])[0] # Logical Block Size (little-endian)
        out['logical_block_size_BE'] =          unpack('>H', pvd[130:132])[0] # Logical Block Size (big-endian) (should be equal to previous)
        out['path_table_size_LE'] =             unpack('<I', pvd[132:136])[0] # Path Table Size (little-endian)
        out['path_table_size_BE'] =             unpack('>I', pvd[136:140])[0] # Path Table Size (big-endian) (should be equal to previous)
        out['location_L_path_table'] =          unpack('<I', pvd[140:144])[0] # Location of Type-L Path Table
        out['location_optional_L_path_table'] = unpack('<I', pvd[144:148])[0] # Location of Optional Type-L Path Table
        out['location_M_path_table'] =          unpack('>I', pvd[148:152])[0] # Location of Type-M Path Table
        out['location_optional_M_path_table'] = unpack('>I', pvd[152:156])[0] # Location of Optional Type-M Path Table
        out['root_directory_entry'] =           pvd[156:190]                  # Directory Entry for Root Directory
        out['volume_set_identifier'] =          pvd[190:318]                  # Volume Set Identifier
        out['publisher_identifier'] =           pvd[318:446]                  # Publisher Identifier
        out['data_preparer_identifier'] =       pvd[446:574]                  # Data Preparer Identifier
        out['application_identifier'] =         pvd[574:702]                  # Application Identifier
        out['copyright_file_identifier'] =      pvd[702:739]                  # Copyright File Identifier
        out['abstract_file_identifier'] =       pvd[739:776]                  # Abstract File Identifier
        out['bibliographic_file_identifier'] =  pvd[776:813]                  # Bibliographic File Identifier
        out['volume_creation_datetime'] =       pvd[813:830]                  # Volume Creation Date and Time
        out['volume_modification_datetime'] =   pvd[830:847]                  # Volume Modification Date and Time
        out['volume_expiration_datetime'] =     pvd[847:864]                  # Volume Expiration Date and Time
        out['volume_effective_datetime'] =      pvd[864:881]                  # Volume Effective Date and Time
        out['file_structure_version'] =         pvd[881]                      # File Structure Version (should always be 1)
        out['offset_882'] =                     pvd[882]                      # should always be 0
        out['application_used'] =               pvd[883:1395]                 # Application Used (not defined by ISO 9660)
        out['reserved'] =                       pvd[1395:2048]                # Reserved by ISO

        # clean strings
        for k in ['identifier', 'system_identifier', 'volume_identifier', 'volume_set_identifier', 'publisher_identifier', 'data_preparer_identifier', 'application_identifier', 'copyright_file_identifier', 'abstract_file_identifier', 'bibliographic_file_identifier']:
            try:
                out[k] = clean_string(out[k])
            except:
                warn("Unable to parse Primary Volume Descriptor '%s' as string: %s" % (k, out[k]))

        # parse date-times
        for k in ['volume_creation_datetime', 'volume_modification_datetime', 'volume_expiration_datetime', 'volume_effective_datetime']:
            try:
                out[k] = IsoFS.parse_pvd_datetime(out[k])
            except:
                warn("Unable to parse PVD '%s' as date-time: %s" % (k, out[k]))

        # parse root directory entry (this must succeed to be able to parse files in this ISO)
        out['root_directory_entry'] = IsoFS.parse_directory_record(out['root_directory_entry'])

        # return final parsed data
        return out

    def parse_volume_descriptor_set_terminator(self):
        '''Return a parsed version of the `Volume Descriptor Set Terminator <https://wiki.osdev.org/ISO_9660#Volume_Descriptor_Set_Terminator>`_ of the ISO.

        Returns:
            `dict`: A parsed version of the Volume Descriptor Set Terminator of the ISO, or `None` if the ISO does not have one.
        '''
        # set things up
        vdst = self.get_volume_descriptor_set_terminator()
        if vdst is None:
            return None
        out = dict()

        # parse raw VDST data
        out['type_code'] =  vdst[0]   # should always be 255
        out['identifier'] = vdst[1:6] # should always be "CD001"
        out['version'] =    vdst[6]   # should always be 1
        out['unused'] =     vdst[7:]  # remaining bytes are not part of ISO 9660

        # clean strings
        for k in ['identifier']:
            try:
                out[k] = clean_string(out[k])
            except:
                warn("Unable to parse Volume Descriptor Set Terminator '%s' as string: %s" % (k, out[k]))

        # return final parsed data
        return out

    def __iter__(self):
        # load root directory entry from PVD
        pvd = self.parse_primary_volume_descriptor()
        if pvd is None:
            return
        to_visit = [(Path(''), pvd['root_directory_entry'])]  # (Path, directory entry) tuples

        # perform search starting from root directory (only contains directories, not files)
        while len(to_visit) != 0:
            # handle current directory
            curr_path, curr_directory_entry = to_visit.pop()
            if curr_path != Path(''):
                yield (curr_path, curr_directory_entry['datetime'], None)

            # read directory data (extent) using ISO LBAs (data_location_LE)
            curr_data = self.read_extent(curr_directory_entry['data_location_LE'], curr_directory_entry['data_length_LE'])
            ind = 0
            while True:
                # load next entry if not at end of this directory
                if ind >= len(curr_data):
                    break
                next_len = curr_data[ind]
                if next_len == 0:
                    break
                next_entry = IsoFS.parse_directory_record(curr_data[ind:ind + next_len])
                next_entry_fn = next_entry['filename']

                # next entry is a directory (add it to `to_visit`)
                if next_entry['file_flags']['is_directory']:
                    if next_entry_fn not in {'', '\x01'}:  # ignore '.' and '..'
                        to_visit.append((curr_path / next_entry_fn, next_entry))

                # next entry is a file (yield it)
                else:
                    next_data = self.read_extent(next_entry['data_location_LE'], next_entry['data_length_LE'])
                    yield (curr_path / next_entry_fn, next_entry['datetime'], next_data)
                ind += next_len



#! /usr/bin/env python
'''
Handle Nintendo GameCube mini-DVD GCM file system
'''

# NiemaFS imports
from niemafs.common import clean_string, FileSystem

# imports
from datetime import datetime
from pathlib import Path
from struct import unpack
from warnings import warn

class GcmFS(FileSystem):
    '''Class to represent a `Nintendo GameCube GCM mini-DVD <https://www.gc-forever.com/yagcd/chap13.html#sec13>`_.'''
    def __init__(self, file_obj, path=None):
        # set things up
        if file_obj is None:
            raise ValueError("file_obj must be a file-like")
        super().__init__(path=path, file_obj=file_obj)
        self.boot_bin = None   # Disk Header (boot.bin)
        self.bi2_bin = None    # Disk Header Information (bi2.bin)
        self.appldr_bin = None # Apploader (appldr.bin)
        self.fst_bin = None    # File System Table (FST) (fst.bin)

        # load header to ensure file validity up-front
        self.get_boot_bin()
        self.get_bi2_bin()
        self.get_appldr_bin()
        self.get_fst_bin()

    def parse_ascii_date(data):
        '''Parse a date in the "YYYY/MM/DD" format.

        Args:
            `data` (`bytes`): A date in the "YYYY/MM/DD" format.

        Returns:
            `datetime`: A Python `datetime` object.
        '''
        if isinstance(data, bytes):
            data = data.decode()
        return datetime.strptime(data, '%Y/%m/%d')

    def get_boot_bin(self):
        '''Return the `Disk Header ("boot.bin") <https://www.gc-forever.com/yagcd/chap13.html#sec13.1>`_ of the GCM.

        Returns:
            `bytes`: The Disk Header ("boot.bin") of the GCM.
        '''
        if self.boot_bin is None:
            self.boot_bin = self.read_file(0x0000, 0x0440)
        return self.boot_bin

    def get_bi2_bin(self):
        '''Return the `Disc Header Information ("bi2.bin") <https://www.gc-forever.com/yagcd/chap13.html#sec13.2>`_ of the GCM.

        Returns:
            `bytes`: The Disc Header Information ("bi2.bin") of the GCM.
        '''
        if self.bi2_bin is None:
            self.bi2_bin = self.read_file(0x0440, 0x2000)
        return self.bi2_bin

    def get_appldr_bin(self):
        '''Return the `Apploader ("appldr.bin") <https://www.gc-forever.com/yagcd/chap13.html#sec13.3>`_ of the GCM.

        Returns:
            `bytes`: The Apploader ("appldr.bin") of the GCM.
        '''
        if self.appldr_bin is None:
            size = unpack('>I', self.read_file(0x2440 + 0x0014, 4))[0]
            self.appldr_bin = self.read_file(0x2440, size)
        return self.appldr_bin

    def get_fst_bin(self):
        '''Return the `File System Table (FST, "fst.bin") <https://www.gc-forever.com/yagcd/chap13.html#sec13.4>`_ of the GCM.

        Returns:
            `bytes`: The File System Table (FST, "fst.bin") of the GCM.
        '''
        if self.fst_bin is None:
            parsed_boot_bin = self.parse_boot_bin()
            self.fst_bin = self.read_file(parsed_boot_bin['fst_offset'], parsed_boot_bin['fst_size'])
        return self.fst_bin

    def parse_boot_bin(self):
        '''Return a parsed version of the `Disk Header ("boot.bin") <https://www.gc-forever.com/yagcd/chap13.html#sec13.1>`_ of the GCM.

        Returns:
            `dict`: A parsed version of the Disk Header ("boot.bin") of the GCM.
        '''
        # set things up
        data = self.get_boot_bin()
        out = dict()

        # parse raw Disk Header (boot.bin) data
        out['game_code'] =             data[0x0000 : 0x0004]                  # Game Code "XYYZ": X = Console ID, YY = Game Code, Z = Country Code
        out['maker_code'] =            data[0x0004 : 0x0006]                  # Maker Code
        out['disk_id'] =               data[0x0006]                           # Disk ID
        out['version'] =               data[0x0007]                           # Version
        out['audio_streaming'] =       data[0x0008]                           # Audio Streaming
        out['stream_buffer_size'] =    data[0x0009]                           # Stream Buffer Size
        out['offsets_0x000A_0x001B'] = data[0x000A : 0x001C]                  # Unused (should be 0s)
        out['dvd_magic_word'] =        data[0x001C : 0x0020]                  # DVD Magic Word (should be 0xc2339f3d)
        out['game_name'] =             data[0x0020 : 0x0400]                  # Game Name
        out['debug_monitor_offset'] =  unpack('>I', data[0x0400 : 0x0404])[0] # Offset of Debug Monitor (dh.bin)?
        out['debug_monitor_address'] = unpack('>I', data[0x0404 : 0x0408])[0] # Address(?) to load Debug Monitor (dh.bin)?
        out['offsets_0x0408_0x0419'] = data[0x0408 : 0x0420]                  # Unused (should be 0s)
        out['main_dol_offset'] =       unpack('>I', data[0x0420 : 0x0424])[0] # Offset of Main Executable Bootfile (main.dol)
        out['fst_offset'] =            unpack('>I', data[0x0424 : 0x0428])[0] # Offset of FST (fst.bin)
        out['fst_size'] =              unpack('>I', data[0x0428 : 0x042C])[0] # Size of FST (fst.bin)
        out['max_fst_size'] =          unpack('>I', data[0x042C : 0x0430])[0] # Max Size of FST (fst.bin) (usually same as previous, except in multi-disc games)
        out['user_position'] =         unpack('>I', data[0x0430 : 0x0434])[0] # User Position(?)
        out['user_length'] =           unpack('>I', data[0x0434 : 0x0438])[0] # User Length(?)
        out['offsets_0x0438_0x043B'] = data[0x0438 : 0x043C]                  # Unknown
        out['offsets_0x043C_0x043F'] = data[0x043C : 0x0440]                  # Unused (should be 0s)

        # clean strings
        for k in ['game_code', 'maker_code', 'game_name']:
            try:
                out[k] = clean_string(out[k])
            except:
                warn("Unable to parse Disk Header (boot.bin) '%s' as string: %s" % (k, out[k]))

        # return final parsed data
        return out

    def parse_bi2_bin(self):
        '''Return a parsed version of the `Disc Header Information ("bi2.bin") <https://www.gc-forever.com/yagcd/chap13.html#sec13.2>`_ of the GCM.

        Returns:
            `dict`: A parsed version of the Disc Header Information ("bi2.bin") of the GCM.
        '''
        # set things up
        data = self.get_bi2_bin()
        out = dict()

        # parse raw Disk Header Information (bi2.bin) data
        out['debug_monitor_size'] =    unpack('>I', data[0x0000 : 0x0004])[0] # Debug-Monitor Size
        out['simulated_memory_size'] = unpack('>I', data[0x0004 : 0x0008])[0] # Simulated Memory Size
        out['argument_offset'] =       unpack('>I', data[0x0008 : 0x000C])[0] # Argument Offset
        out['debug_flag'] =            unpack('>I', data[0x000C : 0x0010])[0] # Debug Flag
        out['track_location'] =        unpack('>I', data[0x0010 : 0x0014])[0] # Track Location
        out['track_size'] =            unpack('>I', data[0x0014 : 0x0018])[0] # Track Size
        out['country_code'] =          unpack('>I', data[0x0018 : 0x001C])[0] # Country Code

        # return final parsed data
        return out

    def parse_appldr_bin(self):
        '''Return a parsed version of the `Apploader ("appldr.bin") <https://www.gc-forever.com/yagcd/chap13.html#sec13.3>`_ of the GCM.

        Returns:
            `dict`: A parsed version of the Apploader ("appldr.bin") of the GCM.
        '''
        # set things up
        data = self.get_appldr_bin()
        out = dict()

        # parse raw Apploader (appldr.bin) data
        out['date'] =                  data[0x0000 : 0x000A]                  # Date (Version) of the Apploader
        out['offsets_0x000A_0x000F'] = data[0x000A : 0x0010]                  # Padding (should be 0s)
        out['apploader_entrypoint'] =  unpack('>I', data[0x0010 : 0x0014])[0] # Apploader Entrypoint
        out['apploader_size'] =        unpack('>I', data[0x0014 : 0x0018])[0] # Apploader Size
        out['trailer_size'] =          unpack('>I', data[0x0018 : 0x001C])[0] # Trailer Size
        out['offsets_0x001C_0x001F'] = data[0x001C : 0x0020]                  # Padding(?)
        out['apploader_code'] =        data[0x0020:]                          # Apploader Code (loaded to 0x81200000 in RAM)

        # parse dates
        for k in ['date']:
            try:
                out[k] = GcmFS.parse_ascii_date(out[k])
            except:
                warn("Unable to parse Apploader (appldr.bin) '%s' as date: %s" % (k, out[k]))

        # return final parsed data
        return out

    def get_string_table(data, string_table_start, num_fst_entries): # TODO DELETE
        '''Get the string table from the File System Table (FST)'s raw data.

        Args:
            `data` (`bytes`): The raw bytes of the FST.

            `string_table_start` (`int`): The start offset of the string table.

            `num_fst_entries` (`int`): The number of entries in the FST.

        Returns:
            `bytes`: The string table.
        '''
        string_table_end = string_table_start
        for i in range(num_fst_entries-1): # FST includes root directory, which isn't in string table
            string_table_end = data.find(b'\00', string_table_end + 1)
        return data[string_table_start : string_table_end + 1] # include the last b'\x00'

    def parse_fst(fst_ind, parent_path, fst, string_table_start):
        '''Recursively parse the `File System Table (FST) <https://www.gc-forever.com/yagcd/chap13.html#sec13.4.1>`_.

        Args:
            `fst_ind` (`int`): This entry's index in the FST.

            `parent_path` (`Path`): The local `Path` of parent of this file/directory, or `None` if this is the root directory.

            `fst` (`bytes`): The raw bytes of the FST.

            `string_table_start` (`int`): Offset in the FST where the string table begins.

        Returns:
            `dict`: The root of the parsed FST.
        '''
        # set things up
        out = {'children':list(), 'is_root':False}

        # parse raw FST entry data
        entry_start_offset = fst_ind * 12
        is_dir =        bool(fst[entry_start_offset])                                                   # Flags: 0 = file, 1 = directory
        st_index =      unpack('>I', b'\x00' + fst[entry_start_offset + 1 : entry_start_offset + 4])[0] # Filename as Offset into String Table
        out['offset'] = unpack('>I', fst[entry_start_offset + 4 : entry_start_offset + 8])[0]           # File Offset (for files) or Parent Offset (for directories)
        out['length'] = unpack('>I', fst[entry_start_offset + 8 : entry_start_offset + 12])[0]          # File Size (for files) or Number of Entries (for root) or Next Offset (for directories

        # determine path
        if parent_path is None: # root
            out['path'] = Path('.')
            out['is_root'] = True
        else: # non-root
            fn_fst_offset = string_table_start + st_index
            fn = fst[fn_fst_offset : fst.find(b'\x00', fn_fst_offset)]
            try:
                fn = fn.decode()
            except:
                warn("Failed to parse filename as string: %s" % fn)
                fn = str(fn)
            out['path'] = parent_path / fn

        # if directory, recursively parse children
        if is_dir:
            fst_ind += 1
            while fst_ind < out['length']:
                child = GcmFS.parse_fst(fst_ind, out['path'], fst, string_table_start)
                fst_ind = child['fst_ind_next']
                out['children'].append(child)
            out['fst_ind_next'] = fst_ind
        else:
            out['fst_ind_next'] = fst_ind + 1

        # return final parsed data
        return out

    def __iter__(self):
        fst = self.get_fst_bin()
        string_table_start = 0x0C * unpack('>I', fst[0x08 : 0x0C])[0]
        to_visit = [GcmFS.parse_fst(0, None, fst, string_table_start)] # start at root directory
        while len(to_visit) != 0:
            file_entry = to_visit.pop()
            if len(file_entry['children']) == 0:
                file_data = self.read_file(file_entry['offset'], file_entry['length'])
            else:
                file_data = None
                to_visit += file_entry['children']
            if not file_entry['is_root']:
                yield (file_entry['path'], None, file_data)



#! /usr/bin/env python
'''
Handle ZIP archives
'''

# NiemaFS imports
from niemafs.common import DEFAULT_BUFFER_SIZE, FileSystem

# standard imports
from datetime import datetime
from pathlib import Path
from warnings import warn

class DirFS(FileSystem):
    '''Class to represent a directory on disk'''
    def __init__(self, path, file_obj=None):
        if file_obj is not None:
            warn("DirFS initializer was given non-None 'file_obj' parameter, which will be ignored")
        super().__init__(path=path, file_obj=None)

    def __iter__(self):
        for curr_path in self.path.rglob('*'):
            curr_timestamp = datetime.fromtimestamp(curr_path.stat().st_mtime)
            if curr_path.is_dir():
                curr_data = None
            else:
                with open(curr_path, mode='rb', buffering=DEFAULT_BUFFER_SIZE) as curr_file:
                    curr_data = curr_file.read()
            yield (curr_path, curr_timestamp, curr_data)



#! /usr/bin/env python
# standard imports
from warnings import warn

# NiemaFS imports
from niemafs.common import clean_string, FileSystem, open_file, safename
from niemafs.dir import DirFS
from niemafs.gcm import GcmFS
from niemafs.iso import IsoFS
from niemafs.zip import ZipFS

# build __all__
__all__ = [
    'clean_string', 'FileSystem', 'open_file', 'safename', # common.py
    'DirFS',                                               # dir.py
    'GcmFS',                                               # gcm.py
    'IsoFS',                                               # iso.py
    'ZipFS',                                               # zip.py
]

# WiiFS depends on PyCryptodome, so import it afterwards
try:
    from niemafs.wii import WiiFS
    __all__.append('WiiFS')
except:
    warn("Unable to import WiiFS. Ensure that you have PyCryptodome installed")